""" projector.py

The `PJLink` class is a connection to a projector using the PJLink protocol.

To provide a "pythonic" API for the different PJLink commands, the
class `CommandGroup` is overriden and groups together related commands.

No state is kept inside the classes.
"""

import os
import re
import asyncio
import hashlib
from enum import Enum


""" Print out messages that are sent and recieved for debugging. """
PRINT_DEBUG_COMMS = bool(os.environ.get("AIOPJLINK_PRINT_DEBUG_COMMS", False))


class PJLinkException(Exception):
    """ Base exception for PJLink library issues. """
    pass


class PJLinkNoConnection(PJLinkException):
    """ Projector did not respond to the connection request. """
    pass


class PJLinkConnectionClosed(PJLinkException):
    """ Projector closed the connection. """
    pass


class PJLinkProtocolError(PJLinkException):
    """ Unexpected communication to or from the projector. """
    pass


class PJLinkUnexpectedResponseParameter(PJLinkException):
    """ Unable to parse a response parameter. """
    pass


class PJLinkPassword(PJLinkException):
    """ Invalid or absent password. """
    pass


class PJLinkProjectorError(PJLinkException):
    """ Projector raised an error when handling a command. """
    pass


class PJLinkERR1(PJLinkProtocolError):
    """ ERR 1, undefined command, as specified in (§2.2)
    """
    pass


class PJLinkERR2(PJLinkException):
    """ ERR 2, out of parameter, as specified in (§2.2)
    """
    pass


class PJLinkERR3(PJLinkException):
    """ ERR 3, unavailable at the current time or in the current projector state, as specified in (§2.2) """
    pass


class PJLinkERR4(PJLinkException):
    """ ERR 4, projector or display failure, as specified in (§2.2) """
    pass


class PJClass(Enum):
    """ Communication protocol message version.

    Class 1 is the most common type of PJLink, and is used for basic commands such as
    power on/off, input selection, and adjusting volume.

    Class 2 is an extended version of the protocol that supports additional commands such
    as opening and closing the projector's lens cover, and is typically used by more sophisticated devices.
    """

    ONE = '1'
    """ PJLink Class 1 command. """

    TWO = '2'
    """ PJLink Class 2 command. """


class PJLink:
    """ Manages a PJLink connection to a projector.

    Usage:

        >>> async with PJLink(address='192.168.100.100', password='secret') as link:
        >>>     await link.power.turn_off()
        >>>     time.sleep(4)
        >>>     await link.power.turn_on()

    """

    C1 = PJClass.ONE
    C2 = PJClass.TWO

    def __init__(self, address, port=4352, password=None, timeout=4, encoding='utf-8'):
        self._reader = None
        self._writer = None
        self._address = address
        self._port = port
        self._encoding = encoding
        self._timeout = timeout
        self._password = password

        # Add the different API namespaces.
        self.info = Information(self)
        self.power = Power(self)
        self.sources = Sources(self)
        self.mute = Mute(self)
        self.errors = Errors(self)
        self.lamps = Lamp(self)
        self.filter = Filter(self)
        self.freeze = Freeze(self)
        self.microphone = Volume(self, 'MVOL')
        self.speaker = Volume(self, 'SVOL')

    async def wait_for_notification(self):
        raise NotImplementedError('class 2 method not supported')

    async def __aenter__(self):
        """ Open a connection to the projector and authenticate. """
        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(self._address, self._port),
                timeout=self._timeout
            )
        except asyncio.exceptions.TimeoutError:
            raise PJLinkNoConnection("timeout - projector did not accept the connection in time")
        except OSError as err:
            raise PJLinkNoConnection(f"os timeout - {str(err)}")

        # An authentication procedure is executed once after each establishment of TCP/IP connection.
        # The authentication procedure involves a password verification process.
        # See https://pjlink.jbmia.or.jp/english/data_cl2/PJLink_5-1.pdf SECTION 5

        # Projector sends first message to identify itself as PJLINK.
        # data = await self._raw_read(n_bytes=9)
        try:
            data = await self._read_next()
        except asyncio.exceptions.TimeoutError:
            raise PJLinkProtocolError('projector did not send a welcome message')
        if len(data) < 9:
            raise PJLinkProtocolError('unexpected opening header message from projector - too short')

        auth_header, auth_enabled, auth_close = data[:7], data[7], data[8]
        if auth_header.upper() != 'PJLINK ':
            raise PJLinkProtocolError('unexpected opening header message from projector - not PJLink')

        # Connection requires no auth: `PJLINK 0\r`.
        if auth_enabled == '0':
            return self

        # Connection requires auth: `PJLINK 1 <token>`.
        if auth_enabled != '1' and auth_close != ' ':
            raise PJLinkProtocolError('unexpected opening security message from projector - unrecognised auth method')

        # Check we have a password specified.
        if self._password is None:
            raise PJLinkPassword('password required')

        # Read the random number used to salt the password (excluding the terminating `\r`).
        token = data[9:-1]
        passcode = (token + self._password).encode('utf-8')
        passcode_md5 = hashlib.md5(passcode).hexdigest()

        # The PJLINK authentication procedure requires the password and the first command to be
        # transmitted together.  We send a power status request for simplicity.
        self._writer.write(bytearray(passcode_md5, encoding=self._encoding))
        self._writer.write(b'%1POWR ?\r')
        await self._writer.drain()

        # Read the first few bytes of the response - check for failed auth.
        # ERRA represents ERR or authorization.
        response = await self._read_next()
        if response.upper() == 'PJLINK ERRA\r':
            raise PJLinkPassword('authentication failed')
        self._parse_response(response, expect_command='POWR', expect_pjclass=PJClass.ONE)
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        """ Close an open connection to the projector. """
        try:
            self._writer.close()
        finally:
            self._reader = None
            self._writer = None

    async def _read_next(self):
        """ Read data until the next terminator (CR) and return the
        message (including CR) as a decoded string."""
        try:
            raw = await asyncio.wait_for(self._reader.readuntil(b'\r'), self._timeout)
        except asyncio.IncompleteReadError as err:
            raise PJLinkConnectionClosed('projector closed the connection') from err
        return raw.decode(self._encoding)

    async def transmit(self, command, param, pjclass: PJClass):
        """ Send a command and get a response.

        Commands and responses are listed here:
        https://pjlink.jbmia.or.jp/english/data_cl2/PJLink_5-1.pdf

        Args:
            command (str): Four letter command (e.g. "POWR").
            param (str): Parameter that accompanies the command (e.g. "?")
            pjclass (str, optional): The PJLink command class. Defaults to '1'.

        Raises:
            PJLinkERR1, PJLinkERR2, PJLinkERR3, PJLinkERR4

        Returns:
            str: The response to the issued command.
        """
        # Generate the command string.
        cstring = self._format_command(command, param, pjclass)

        # Send the command string.
        cbytes = bytearray(cstring, self._encoding)
        if PRINT_DEBUG_COMMS:
            print("🚢", cbytes)
        self._writer.write(cbytes)
        await self._writer.drain()

        # Get the response.
        response = await self._read_next()

        # Parse the response.
        _, param = PJLink._parse_response(response, expect_command=command, expect_pjclass=pjclass)
        return param

    @staticmethod
    def _format_command(command, param, pjclass: PJClass):
        pjclass = PJClass(pjclass)
        if not command.isupper():
            raise PJLinkProtocolError('command is not uppercase')
        if len(command) != 4:
            raise PJLinkProtocolError('command is not 4 bytes')
        if len(param) > 128:
            raise PJLinkProtocolError('command param is larger than 128 bytes')
        sep = ' '
        return f'%{pjclass.value}{command}{sep}{param}\r'

    @staticmethod
    def _parse_response(data, expect_command=None, expect_pjclass=PJClass.ONE):
        # NOTE: Postels robustness principle - be conservative in what you do, be liberal in what you accept from others
        if PRINT_DEBUG_COMMS:
            print('➡️ ', data)
        expect_pjclass = PJClass(expect_pjclass)

        # Check header and class version.
        header, version = data[0], data[1]
        if header != '%':
            raise PJLinkProtocolError('unexpected response header')
        if version != expect_pjclass.value:
            raise PJLinkProtocolError('unexpected response protocol class')

        # Grab the command body, separator, and param.
        command = f'{data[2:6]}'.upper()
        sep = data[6]
        param = data[7:-1]

        # Check them for correctness.
        if sep != '=':
            raise PJLinkProtocolError('unexpected response separator')
        if expect_command is not None and command != expect_command:
            raise PJLinkProtocolError('unexpected response command')

        # Handle for protocol and projector errors.
        param_u = param.upper()
        if param_u == 'ERR1':
            raise PJLinkERR1('unsupported command')
        elif param_u == 'ERR2':
            raise PJLinkERR2('out of parameter')
        elif param_u == 'ERR3':
            raise PJLinkERR3('unavailable in the current state')
        elif param_u == 'ERR4':
            raise PJLinkERR4('projector or display failure')

        return command, param


class CommandGroup:
    """ Base class for related groups of PJLink functionality.
    """

    def __init__(self, link: PJLink):
        self._link = link

    async def _transmit_ok(self, command, param, pjclass):
        """ Transmit a command and check the response is OK. """
        response = await self._link.transmit(command, param, pjclass)
        if response.upper() != 'OK':
            raise PJLinkUnexpectedResponseParameter('expected OK response')


class Power(CommandGroup):
    """ Control and query the power state of the projector lamp. """

    class State(Enum):
        """ PJLink projector lamp states (combining §4.1 and §4.2). """
        OFF = '0'
        ON = '1'
        COOLING = '2'
        WARMING = '3'

        def __bool__(self):
            """ Truthy states for `on` and `warming`, falsy states for `off` and `cooling`. """
            return self in (Power.State.ON, Power.State.WARMING)

    ON = State.ON
    OFF = State.OFF

    async def set(self, state: State, pjclass=PJClass.ONE):
        """ Send a power control instruction to power the projector lamp on or off. """
        state = Power.State(state)
        if state in (Power.State.COOLING, Power.State.WARMING):
            raise ValueError('expected Power.State.ON or Power.State.OFF')
        await self._transmit_ok('POWR', state.value, pjclass=pjclass)

    async def get(self, pjclass=PJClass.ONE):
        """ Request the power status of the projector. """
        return Power.State(await self._link.transmit('POWR', '?', pjclass=pjclass))

    async def turn_on(self):
        """ Power the projector on. """
        await self.set(Power.State.ON)

    async def turn_off(self):
        """ Power the projector off. """
        await self.set(Power.State.OFF)


class Sources(CommandGroup):
    """ Control and query projector input sources. """

    class Mode(Enum):
        """ Display input source modes (§4.3)."""
        RGB = '1'
        VIDEO = '2'
        DIGITAL = '3'
        STORAGE = '4'
        NETWORK = '5'
        INTERNAL = '6'
        """ Class 2 only. """

    async def set(self, mode, index, pjclass=PJClass.ONE):
        """ Get the current source input selection (§4.3).
        """
        mode, index = self._check_mode_index(mode, index)
        return await self._transmit_ok(command='INPT', param=f'{mode.value}{index}', pjclass=pjclass)

    async def get(self, pjclass=PJClass.ONE):
        """ Get the current source input selection (§4.4).
        """
        values = await self._link.transmit(command='INPT', param='?', pjclass=pjclass)
        if len(values) != 2:
            raise PJLinkUnexpectedResponseParameter('expected 2 INPT response characters')
        return Sources.Mode(values[0]), values[1]

    def _check_mode_index(self, mode, index):
        """ Ensure a mode and index pair are valid. """
        mode = Sources.Mode(mode)
        index = f'{index}'
        if len(index) != 1:
            raise ValueError('index must be a single character (1-9 for Class 1, and 1-9A-Z for Class 2)')
        return mode, index

    async def available(self, pjclass=PJClass.ONE):
        """ List all the available input sources (§4.9).

        Returns:
            A list of available input sources in the format: (Sources.Mode, index str).  For example:
                [(<Mode.RGB: '1'>, '1'), (<Mode.DIGITAL: '3'>, '1'), ...]
        """
        sources = await self._link.transmit('INST', '?', pjclass)
        sources = sources.split(' ')
        return [(Sources.Mode(first), second) for first, second in sources]

    async def get_source_name(self, mode, index):
        """ Get the name of a given input source (§4.17).
        :param mode Sources.Mode: The input mode to select.
        :param index str: A single character.
        """
        mode, index = self._check_mode_index(mode, index)
        return await self._link.transmit(command='INNM', param=f'?{mode.value}{index}', pjclass='2')

    async def available_with_names(self):
        """ List all the available input sources with names (§4.17).
        Returns:
            A list of available input sources in the format (Source.Mode, index str, name str).
            For example:
                [(<Mode.RGB: '1'>, '1', 'Computer'), (<Mode.DIGITAL: '3'>, '1', 'DVI-D'), ...]
        """
        sources = await self.available(pjclass=PJClass.TWO)
        output = []
        for mode, index in sources:
            try:
                name = await self.get_source_name(mode, index)
            except PJLinkERR2:
                name = None
            output.append((mode, index, name))
        return output

    async def resolution(self):
        """ Get the current projector resolution (§4.18)
        Returns:
            (x:int, y:int) tuple: Horizontal and vertical resolutions of input signal respectively.
        """
        response = await self._link.transmit('IRES', '?', PJClass.TWO)
        if response == '-':
            raise PJLinkProjectorError('no signal input')
        if response == '*':
            raise PJLinkProjectorError('unknown signal')

        # Convert each axis to an integer.
        try:
            resolution = [int(dim) for dim in re.split('x', response, flags=re.IGNORECASE)]
            return tuple(resolution)
        except ValueError:
            raise PJLinkUnexpectedResponseParameter('unable to parse resolution')

    async def recommended_resolution(self):
        """ Get the current recommended resolution (§4.19)
        Returns:
            (x, y) tuple: Horizontal and vertical resolutions of input signal respectively.
        """
        response = await self._link.transmit('RRES', '?', PJClass.TWO)
        try:
            resolution = [int(dim) for dim in re.split('x', response, flags=re.IGNORECASE)]
            return tuple(resolution)
        except ValueError:
            raise PJLinkUnexpectedResponseParameter('unable to parse resolution')


class Mute(CommandGroup):
    """ Control audio and visual track mute status (§4.5, §4.6).

    If the mute function is individually executed or cancelled for the models
    that do not have audio or video mute functions, "ERR 2" (out of parameter range) is returned.
    """

    async def status(self):
        """ Current (video, audio) track mute status returned as two booleans (§4.6).

        Returns:
            tuple(video: bool, audio: bool): True if the track is muted.  False if not.
        """
        status = await self._link.transmit('AVMT', '?', pjclass=PJClass.ONE)
        if status == '11':
            return True, False
        elif status == '21':
            return False, True
        elif status == '31':
            return True, True
        elif status == '30':
            return False, False
        else:
            raise PJLinkUnexpectedResponseParameter('unexpected mute response')

    async def video(self, muted: bool):
        """ Set if the video track should be muted (True to mute, False to unmute). """
        cmd = '1' if muted is True else '0'
        await self._transmit_ok('AVMT', f'1{cmd}', pjclass=PJClass.ONE)

    async def audio(self, muted: bool):
        """ Set if the audio track should be muted (True to mute, False to unmute). """
        cmd = '1' if muted is True else '0'
        await self._transmit_ok('AVMT', f'2{cmd}', pjclass=PJClass.ONE)

    async def both(self, muted: bool):
        """ Set if the AV tracks should be muted (True to mute, False to unmute). """
        cmd = '1' if muted is True else '0'
        await self._transmit_ok('AVMT', f'3{cmd}', pjclass=PJClass.ONE)

    async def set(self, video: bool, audio: bool):
        """ Enable or disable mute for each track (call mirrors output of `status`).
        :param video (bool): True to mute. False to unmute. None to skip.
        :param audio (bool): True to mute. False to unmutes. None to skip.
        """
        # Skip non-specified condition.
        if video is None and audio is None:
            return

        # Fully specified conditions.
        if video is True and audio is True:
            await self.both(True)
        elif video is False and audio is False:
            await self.both(False)

        # Partially specified conditions.
        else:
            if video is not None:
                await self.video(video)
            if audio is not None:
                await self.audio(audio)


class Errors(CommandGroup):
    """ Provide information about errors occuring within the projector (§4.7).
    """
    class Category(Enum):
        """ The different types of error returned according to (§4.7). """
        FAN = 'fan'
        LAMP = 'lamp'
        TEMP = 'temperature'
        COVER = 'cover'
        FILTER = 'filter'
        OTHER = 'other'

    class Level(Enum):
        """ Error level for each `Category` (§4.7). """
        OK = '0'
        WARN = '1'
        ERROR = '2'

    async def query(self):
        """ Query the projecteor for the latest error status
        information for each of the error categories (§4.7).

        Returns:
            dict[Category]: Level: Table of error categories to states.
        """
        errors = await self._link.transmit('ERST', '?', pjclass=PJClass.ONE)
        try:
            fan, lamp, temp, cover, filt, other = errors
        except ValueError:
            raise PJLinkUnexpectedResponseParameter('unexpected number of error types reported')
        try:
            return {
                Errors.Category.FAN: Errors.Level(fan),
                Errors.Category.LAMP: Errors.Level(lamp),
                Errors.Category.TEMP: Errors.Level(temp),
                Errors.Category.COVER: Errors.Level(cover),
                Errors.Category.FILTER: Errors.Level(filt),
                Errors.Category.OTHER: Errors.Level(other),
            }
        except ValueError as err:
            raise PJLinkUnexpectedResponseParameter('unknown error level') from err


class Lamp(CommandGroup):
    """ Status information about the projector light sources (§4.8).

        According to the spec the "usage time of lamp is always 0 when it is
        not counted by the projector."
    """
    class State(Enum):
        OFF = '0'
        ON = '1'

    async def status(self):
        """ Query the current lamp hours and lamp statuses.
        There may be more than one lamp in some projectors, so this is returned
        as a list.
        Returns:
            [(hours:int, state:Lamp.State)]: List of lamp hours and states for each lamp.
        """
        # Express a special meaning for ERR1 (§4.8).
        try:
            response = await self._link.transmit('LAMP', '?', pjclass=PJClass.ONE)
        except PJLinkERR1:
            raise PJLinkERR1('no lamp')

        # Split the response by " " and then pair up all the numbers from
        # the list in groups of 2.  If there are any remainders, raise an error.
        # See: https://docs.python.org/3/library/itertools.html (grouper)
        # This takes a line like: "1000 1 50 0" and breaks it into pairs:
        #   (1000, 1), (50, 0)
        # Such that these can then be remapped into our high level interface.
        try:
            # Python 3.9 solution.
            pairs = []
            numbers = response.split(' ')
            for i in range(0, len(numbers), 2):
                pairs.append(numbers[i:i + 2])

            # Python 3.10 solution.
            # pairs = zip(*[iter(response.split(' '))] * 2, strict=True)
            # pairs = [pair for pair in pairs]#

            # Remap the statuses to integers and our lamp state enum.
            return [(int(hours), Lamp.State(state)) for hours, state in pairs]
        except ValueError as err:
            raise PJLinkUnexpectedResponseParameter('unparsable lamp status') from err

    async def hours(self):
        """ How long has the first lamp been on (hour, integer).

        According to the spec (§4.8) the "usage time of lamp is always 0 when it is
        not counted by the projector."
        """
        return (await self.status())[0][0]

    async def replacement_models(self):
        """ Get the lamp replacement models listed in the projector.
        There may be more than one model number, so they are returned in a list.
        """
        models = await self._link.transmit('RLMP', '?', pjclass=PJClass.TWO)
        return [m for m in models.split(' ') if m]


class Filter(CommandGroup):
    """ Status information about the projector filters (§4.20, §4.22).
    """

    async def hours(self):
        """ Query the filter usage time (§4.20).
        Filter usage time is always 0 when it is not counted by the projector.
        """
        # Request the value.
        try:
            return int(await self._link.transmit('FILT', '?', pjclass=PJClass.TWO))

        # Express a special meaning for ERR1 (§4.20).
        except PJLinkERR1:
            raise PJLinkERR1('no filter')

        # Parse issue.
        except ValueError:
            raise PJLinkUnexpectedResponseParameter('filter usage not parsable')

    async def replacement_models(self):
        """ Get the filter replacement models listed in the projector (§4.22).
        There may be more than one model number, so they are returned in a list.
        """
        models = await self._link.transmit('RFIL', '?', pjclass=PJClass.TWO)
        return [m for m in models.split(' ') if m]


class Freeze(CommandGroup):
    """ Controls freezing and unfreezing the current frame (§4.25, §4.26).
    """

    async def set(self, freeze: bool):
        """ Freeze or unfreeze the screen (§4.25). """
        cmd = '1' if bool(freeze) else '0'
        await self._transmit_ok('FREZ', cmd, pjclass=PJClass.TWO)

    async def get(self):
        """ Returns True if the screen is currently frozen, and False if not §4.26. """
        response = await self._link.transmit('FREZ', '?', pjclass=PJClass.TWO)
        if response == '0':
            return False
        elif response == '1':
            return True
        else:
            raise PJLinkUnexpectedResponseParameter('unexpected freeze state')


class Volume(CommandGroup):
    """ Controls a xVOL style command (e.g. for speakers and microphones) as
    defined in (§4.23, §4.24).

    According to the spec:
        "As for a specification to increase the microphone volume by one level when it
        is in the maximum state, and a specification to decrease the microphone
        volume by one level when it is in the minimum state, the response
        for a normal case is returned."

    Volume related to audio output (audio out, built-in speaker in equipment
    model, etc.) is referred to as the speaker volume.

    Volume related to voice input (audio in, microphone terminal to be input
    to the model, etc.) is referred to as the microphone volume.
    """

    def __init__(self, link, instruction):
        super().__init__(link)
        self.instruction = instruction

    async def turn_up(self):
        """ Increase the volume by one unit. """
        await self._transmit_ok(self.instruction, '1', pjclass=PJClass.TWO)

    async def turn_down(self):
        """ Decrease the volume by one unit. """
        await self._transmit_ok(self.instruction, '0', pjclass=PJClass.TWO)


class Information(CommandGroup):
    """ Gathers information about the projector.
    """

    async def table(self):
        """ Collect a table of all the different information available
        from this projector.  If the projector responds, an empty string is
        returned, but if it throws an error, `None` is returned.

        See the code for the dictionary entries.
        """

        # Helper to ensure it is always returned regardless of the exception.
        async def _safe(method):
            try:
                return await method()
            except Exception:
                return None

        # Table.
        return {
            "software_version": await _safe(self.software_version),
            "serial_number": await _safe(self.serial_number),
            "pjlink_class": await _safe(self.pjlink_class),
            "other": await _safe(self.other),
            "product_name": await _safe(self.product_name),
            "manufacturer_name": await _safe(self.manufacturer_name),
            "projector_name": await _safe(self.projector_name),
        }

    async def software_version(self):
        """ Request software version of the projector (§4.16).
        The version information of the software defined by the manufacturer is indicated.
        Version information can be expressed in any way.

        Returns:
            str: The version string.
        """
        return await self._link.transmit('SVER', '?', PJClass.TWO)

    async def serial_number(self):
        """ Request the projector serial number (§4.15).
        The serial number information defined by the manufacturer is indicated.

        Returns:
            str: The serial number string.
        """
        return await self._link.transmit('SNUM', '?', PJClass.TWO)

    async def pjlink_class(self, pjclass=PJClass.ONE):
        """ Get projectors PJLink class number as a `PJClass` enumeration (§4.14) """
        try:
            return PJClass(await self._link.transmit('CLSS', '?', pjclass=pjclass))
        except ValueError as err:
            raise PJLinkUnexpectedResponseParameter('unexpected PJLink class') from err

    async def other(self):
        """ Query the projector for other information about the projector/display
        described by the manufacture. Defined as in (§4.13).

        If there is no other information, this returns an empty string.
        """
        return await self._link.transmit('INFO', '?', PJClass.ONE)

    async def product_name(self):
        """ Get product name information string (e.g. EPSON PU1007B/PU1007W) as in (§4.12).

        If there is no information, this returns an empty string.
        """
        return await self._link.transmit('INF2', '?', PJClass.ONE)

    async def manufacturer_name(self):
        """ Get manufacturer name information string (e.g. EPSON) as in (§4.11).

        If there is no information, this returns an empty string.
        """
        return await self._link.transmit('INF1', '?', PJClass.ONE)

    async def projector_name(self):
        """ Get projector name information string (e.g. EBB13648) as in (§4.10).

        If there is no information, this returns an empty string.
        """
        return await self._link.transmit('NAME', '?', PJClass.ONE)



""" Basic CLI for testing.
"""

import sys
import asyncio
from aiopjlink.projector import PJLink


async def cli():

    # Parse
    ip = sys.argv[1] if len(sys.argv) >= 2 else "192.168.1.120"
    cmd = sys.argv[2] if len(sys.argv) >= 3 else "off"
    cmd = cmd.lower()

    # Send the mssage.
    async with PJLink(address=ip) as link:

        if cmd == "on":
            await link.power.turn_on()
            print(ip, "on")

        if cmd == "off":
            await link.power.turn_off()
            print(ip, "off")

        if cmd == "errors":
            print(ip, "errors = ", await link.errors.query())


if __name__ == '__main__':
    loop = asyncio.new_event_loop()
    loop.run_until_complete(cli())



import asyncio
import hashlib
import unittest
import platform
import contextlib

import aiopjlink.projector as aiopjlink


# Allow the loop to close cleanly on Windows.
if platform.system() == "Windows":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class PJLinkTestFrameworkError(Exception):
    """ Base exception for problems with the test framework. """
    pass


class NoClientMessage(PJLinkTestFrameworkError):
    """ The test case fails because the test case didn't send a client message. """
    pass


class UnexpectedClientMessage(PJLinkTestFrameworkError):
    """ The test case fails because the test case expected something the client didn't send. """
    pass


class NoExpectations(PJLinkTestFrameworkError):
    """ The test case fails because the mock projector recieves a command before it is told how to respond. """
    pass


class PJLinkServerProtocol(asyncio.Protocol):
    """
    Provides a method for mocking a projector, including a full socket
    stack to provide a proper simulation of a projector for the client - including
    erroneous behaviours.

    This class has some nuance - check it here.
    https://docs.python.org/3/library/asyncio-protocol.html
    """

    def __init__(self, loop, debug=False):

        # Event loop of the current test case.
        self.loop = loop

        # Should send and recieve messages be printed.
        self.debug = debug

        # An opening message is sent as soon as a connection is established.
        self._opening_message = None

        # What the projector should expect to recieve / do next (as defined by the test case).
        self._expected_events = []

        # Buffer fore incoming messages until \r.
        self._recv_buffer = b''

    def _write(self, data):
        """ Send data to the client. """
        if self.debug:
            print("PJLinkServerProtocol SEND:", data)
        self.transport.write(data)

    def connection_made(self, transport):
        """ Called when a connection is first made to this projector. """
        self.transport = transport
        if self.debug:
            peer = self.transport.get_extra_info('peername')
            print("PJLinkServerProtocol CONNECTION_MADE:", peer)
        if self._opening_message:
            self._write(self._opening_message)

    def connection_lost(self, exc):
        """ Called when the connection is lost or closed. """
        if self.debug:
            peer = self.transport.get_extra_info('peername')
            print("PJLinkServerProtocol CONNECTION_LOST:", peer)

    def data_received(self, data):
        """ Called when the projector recieves data from the client. """
        if self.debug:
            print("PJLinkServerProtocol RECV:", data)

        # Buffer up writes until we get a terminator.
        self._recv_buffer += data
        if self._recv_buffer.endswith(b'\r'):

            # Handle bad test case programming.
            if not len(self._expected_events):
                raise NoExpectations('mock projector got a message before it was told to expect data')

            # Pop the next expected event off.
            expected = self._expected_events.pop(0)

            # If the client sent an _unexpected_ message, flag it as incorrect
            # then save the contents of the buffer, and close the connection.
            if self._recv_buffer != expected.incoming:
                if self.debug:
                    print("PJLinkServerProtocol UNEXPECTED:", self._recv_buffer, 'expected', expected.incoming)
                expected.recv_buffer_contents = self._recv_buffer[:]
                self.loop.call_soon_threadsafe(expected.set)
                self._recv_buffer = b''
                self.transport.close()

            # If the client sent an _expected_ message, flag it as correct,
            # save the buffer contents, and then reply with the expected
            # response.
            else:
                expected.recv_buffer_contents = self._recv_buffer[:]
                self.loop.call_soon_threadsafe(expected.set)
                self.transport.write(expected.respond_with)
                self._recv_buffer = b''

    def open_and_send(self, message):
        """ Set the message to send to the client when a connection is first established. """
        self._opening_message = message

    @contextlib.asynccontextmanager
    async def when(self, incoming, respond_with, within=1):
        """
        When the projector recieves an incoming message, it should respond with
        a reply within n seconds.

        If the projector recieves nothing from the client, raise `NoClientMessage`.
        If the projector recieves a different message, raise `UnexpectedClientMessage`.

        Other exceptions are passed through for handling by the test case.
        """
        # Define the expected incoming message and response.
        expected = ExpectedEvent(incoming, respond_with, within=within)
        self._expected_events.append(expected)

        # Give the event back to the application.
        try:
            yield expected
            await asyncio.wait_for(expected.wait(), timeout=expected.timeout)

        # Our mock protocol closes the server if it finds an error.
        # This results in the client seeing a disconnection (it doesn't know its mocked)
        # so we check the result of our event.

        # Handle reasons our tests might not be written correctly.
        except asyncio.exceptions.TimeoutError:
            # print("⌚ probably no message from client")
            if not expected.recv_buffer_contents:
                raise NoClientMessage(f'projector recieved no data from client (within={expected.timeout}s)')

        except aiopjlink.PJLinkConnectionClosed:
            # print("🚌 probably transport closed on purpose after bad message")
            # Check the object state to see if it got a bad message.
            if expected.recv_buffer_contents != expected.incoming:
                raise UnexpectedClientMessage('projector expected {!r} from the client but got {!r}'.format(expected.incoming, expected.recv_buffer_contents)) # noqa

        # Ensure the event is removed.
        # Handles the edge case where the client API raises an exception before transmission.
        finally:
            # # Check the object state to see if it ever got a message from the client.
            # print("📋 expected.incoming", expected.incoming)
            # print("📋 expected.recv_buffer_contents", expected.recv_buffer_contents)

            if expected in self._expected_events:
                self._expected_events.remove(expected)


class ExpectedEvent(asyncio.Event):
    """
    Specalised event that controls what the mocked projector
    does next and validates that it behaves in the expected way.
    """
    def __init__(self, incoming, respond_with, within):
        super().__init__()
        # What the projector expects to recieve from the client.
        self.incoming = incoming

        # What should the projector respond with when it receives the expected
        # message from the client. NOTE: If it doesn't get the expected message
        # it hangs up the connection.  In this way, simulated projector errors
        # need to be implemented by the test cases.
        self.respond_with = respond_with

        # How long does the client have to do its thing.
        self.timeout = within

        # Contents of the projector recieve buffer after processing.
        self.recv_buffer_contents = None


@contextlib.asynccontextmanager
async def mock_tcp_pjlink(host='127.0.0.1', port=4352, password=None):
    loop = asyncio.get_running_loop()
    protocol = PJLinkServerProtocol(loop=loop)
    server = await loop.create_server(lambda: protocol, host, port)
    try:
        server_task = loop.create_task(server.serve_forever())
        yield protocol
    finally:
        server_task.cancel()


@contextlib.asynccontextmanager
async def mock_client_server_noauth():
    async with mock_tcp_pjlink() as server:
        server.open_and_send(b'PJLINK 0\r')
        async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:
            yield server, client


class ReflectiveMockTests(unittest.IsolatedAsyncioTestCase):
    """ Validate the test case framework. """

    async def test_mock_nonresponse(self):
        """
        The test framework notices that the client has not issued the expected command.
        It times out with a `NoClientMessage` and closes the server.
        """
        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')

            # Expect a POWR status request.
            with self.assertRaises(NoClientMessage):
                async with server.when('%1POWR ?', respond_with=b'%1POWR=0\r'):
                    # Never give one.
                    pass

    async def test_mock_unexpectedresponse(self):
        """
        The test framework notices that the client has issued an "incorrect" command
        to the one that was expected (in this case, during auth).
        It catches this error, closes the server, and issues an `UnexpectedClientMessage`.

        This helps us check the tests are written correctly without super complex error messages.
        """
        # Start server with auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 1 21d0e96e\r')

            # Expect an unexpected message (valid test behaviour).
            with self.assertRaises(UnexpectedClientMessage):

                # Compute what the projector expects to see if the password was actually "ABC123"
                # with the above salt token (21d0e96e).
                salted_password = hashlib.md5('21d0e96eABC123'.encode()).hexdigest().encode()
                expected_cmd = bytes(salted_password)+b'%1POWR ?\r'

                # Tell the test framework to raise an UnexpectedClientMessage
                # if we get a different password sent from the client.  In this case, that is
                # what we want (not an auth error) because we are testing our ability to mock
                # a projector and NOT the behaviour of the projector.
                async with server.when(expected_cmd, respond_with=b'%1POWR=0\r'):

                    # Give a junk password to trigger the UnexpectedClientMessage.
                    async with aiopjlink.PJLink(address='127.0.0.1', password='INCORRECT'):
                        pass

    async def test_mock_unexpectedresponse_after_connection(self):
        """
        The test framework allows: (1) successful messages to be handled OK, (2) successive
        successful messages, and (3) catches errors in successive unexpected messages.

        This ensures the test framework is suitable for multiple calls (e.g. internal buffers
        are reset, and that sort of thing).
        """
        # Open a connection successfully with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')
            async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:

                # Check that the test framework recieves the expected message.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=0\r'):
                    value = await client.transmit('POWR', '?', pjclass=aiopjlink.PJClass.ONE)
                    self.assertEqual(value, '0')

                # Check that the test framework recieves the expected message - 2nd time.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=0\r'):
                    value = await client.transmit('POWR', '?', pjclass=aiopjlink.PJClass.ONE)
                    self.assertEqual(value, '0')

                # Check that the test framework handles the mistake in the test code.
                with self.assertRaises(UnexpectedClientMessage):
                    async with server.when(b'SOME_MESSAGE\r', respond_with=b'%1POWR=0\r'):
                        await client.transmit('POWR', '?', pjclass=aiopjlink.PJClass.ONE)

    async def test_mock_response_stacking(self):
        """ The test framework allows expected responses to be queued.
        """
        # Open a connection successfully with no auth.
        async with mock_client_server_noauth() as (server, client):

            # Expect a POWR followed by a CLSS.
            async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=0\r'):
                async with server.when(b'%1CLSS ?\r', respond_with=b'%1CLSS=2\r'):
                    value = await client.transmit('POWR', '?', pjclass=aiopjlink.PJClass.ONE)
                    self.assertEqual(value, '0')
                    value = await client.transmit('CLSS', '?', pjclass=aiopjlink.PJClass.ONE)
                    self.assertEqual(value, '2')

            # Expect a POWR followed by a CLSS - but don't get it, so we should
            # expect a warning about the test case: UnexpectedClientMessage
            with self.assertRaises(UnexpectedClientMessage):
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=0\r'):
                    async with server.when(b'%1CLSS ?\r', respond_with=b'%1CLSS=2\r'):
                        await client.transmit('CLSS', '?', pjclass=aiopjlink.PJClass.ONE)
                        await client.transmit('POWR', '?', pjclass=aiopjlink.PJClass.ONE)


class AuthTests(unittest.IsolatedAsyncioTestCase):
    """ PJLink authentication behaves as expected.
    """

    async def test_auth_none(self):
        """ Tests a connection with no authentication. """

        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')

            # Prime the server to handle a power request.
            async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=0\r'):

                # Send the power request and check the response is expected.
                async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:
                    value = await client.transmit('POWR', '?', pjclass='1')
                    self.assertEqual(value, '0')

    async def test_auth_malformed(self):
        """ Tests the projector sending back a malformed welcome message generates a `PJLinkProtocolError`. """

        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK XXX\r')

            # Expect to see a protocol error when we connect.
            with self.assertRaises(aiopjlink.PJLinkProtocolError):
                async with aiopjlink.PJLink(address='127.0.0.1', password=None):
                    pass

    async def test_auth_no_welcome(self):
        """ Tests the projector not sending a welcome message generates a `PJLinkProtocolError` """

        # Start server with no auth.
        async with mock_tcp_pjlink():

            # Do not send a message (i.e. the one commented out below).
            # server.open_and_send(b'PJLINK XXX\r')

            # Expect to see a protocol error when we connect.
            with self.assertRaises(aiopjlink.PJLinkProtocolError):
                async with aiopjlink.PJLink(address='127.0.0.1', password=None, timeout=0.5):
                    pass

    async def test_auth_no_server(self):
        """ Tests that the client honours the timeout if no server responds to the connection. """

        # Expect to see a PJLinkNoConnection when we connect.
        # CONDITION 1: The host can be reached by the OS but no response (aiotimeout).
        with self.assertRaises(aiopjlink.PJLinkNoConnection) as err:
            async with aiopjlink.PJLink(address='127.0.0.1', password=None, timeout=0.5):
                pass
        self.assertEqual(str(err.exception), 'timeout - projector did not accept the connection in time')

        # CONDITION 2: The host cannot be reached by the OS.
        with self.assertRaises(aiopjlink.PJLinkNoConnection) as err:
            async with aiopjlink.PJLink(address='0.0.0.0', password=None, timeout=0.5):
                pass
        self.assertIn('os timeout', str(err.exception))

    async def test_auth_valid_pw(self):
        """ Tests a connection with valid authentication. """

        # Server sends the auth challenge: token='21d0e96e'; password='abc123'
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 1 21d0e96e\r')

            # Calcuate the expected projector behaviour.
            # NOTE: Our client implementation always sends a `POWR ?` since that is
            # a commonly implemented command.
            salted_password = hashlib.md5('21d0e96eabc123'.encode()).hexdigest().encode()
            cmd = bytes(salted_password)+b'%1POWR ?\r'
            async with server.when(cmd, respond_with=b'%1POWR=0\r'):

                # Send the power request and check the response is expected.
                async with aiopjlink.PJLink(address='127.0.0.1', password='abc123'):
                    pass

    async def test_auth_invalid_pw(self):
        """ Tests a connection with invalid authentication. """

        # Server sends the auth challenge: token='21d0e96e'; password='abc123'
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 1 21d0e96e\r')

            # Calcuate the expected projector behaviour.
            salted_password = hashlib.md5('21d0e96eINVALIDPW'.encode()).hexdigest().encode()
            cmd = bytes(salted_password)+b'%1POWR ?\r'
            async with server.when(cmd, respond_with=b'PJLINK ERRA\r'):

                # Send the power request and check the response is expected.
                with self.assertRaises(aiopjlink.PJLinkPassword):
                    async with aiopjlink.PJLink(address='127.0.0.1', password='INVALIDPW'):
                        pass


class ProtocolTests(unittest.IsolatedAsyncioTestCase):
    """ PJLink protocol managment behaves as expected.
    """

    async def test_response_parsing_high_level(self):
        """ Check that responses are handled correctly by the client. """

        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')
            async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:

                # Unxpected command sent as a result of a request.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1ROWP=A\r'):
                    with self.assertRaises(aiopjlink.PJLinkProtocolError):
                        await client.power.get()

    async def test_command_construction(self):
        """ Test that command formatting accepts valid values and raises errors if out of spec. """
        # Normal commands.
        result = aiopjlink.PJLink._format_command('ABCD', '???', pjclass='1')
        self.assertEqual(result, '%1ABCD ???\r')

        result = aiopjlink.PJLink._format_command('EFGH', '1', pjclass='1')
        self.assertEqual(result, '%1EFGH 1\r')

        result = aiopjlink.PJLink._format_command('IJKL', '9999999', pjclass='1')
        self.assertEqual(result, '%1IJKL 9999999\r')

        result = aiopjlink.PJLink._format_command('ABCD', '?', pjclass='2')
        self.assertEqual(result, '%2ABCD ?\r')

        # Bad commands.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            result = aiopjlink.PJLink._format_command('abcd', '?', pjclass='1')
        self.assertEqual(str(err.exception), 'command is not uppercase')

        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            result = aiopjlink.PJLink._format_command('ABCDE', '?', pjclass='1')
        self.assertEqual(str(err.exception), 'command is not 4 bytes')

        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            large_param = 'X' * 129
            result = aiopjlink.PJLink._format_command('ABCD', large_param, pjclass='1')
        self.assertEqual(str(err.exception), 'command param is larger than 128 bytes')

        with self.assertRaises(ValueError) as err:
            result = aiopjlink.PJLink._format_command('ABCD', '?', pjclass='3')
        self.assertEqual(str(err.exception), '\'3\' is not a valid PJClass')

    async def test_response_parsing(self):
        """ Ensure that errors generate the correct responses. """

        # Valid data.
        command, parameter = aiopjlink.PJLink._parse_response(
            data='%1ABCD=5\r',
            expect_command='ABCD',
            expect_pjclass='1')
        self.assertEqual(command, 'ABCD')
        self.assertEqual(parameter, '5')

        # Bad response header.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            aiopjlink.PJLink._parse_response(
                data='#1ABCD=5\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unexpected response header')

        # Bad class version.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            aiopjlink.PJLink._parse_response(
                data='%2ABCD=5\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unexpected response protocol class')

        # Bad class version.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            aiopjlink.PJLink._parse_response(
                data='%2ABCD=5\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unexpected response protocol class')

        # Bad separator.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCD/5\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unexpected response separator')

        # Bad command length.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCDE/5\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unexpected response separator')

        # Bad command length.
        with self.assertRaises(aiopjlink.PJLinkProtocolError) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCD=5\r',
                expect_command='XXXX',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unexpected response command')

        # ERR1 - unsupported command
        with self.assertRaises(aiopjlink.PJLinkERR1) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCD=ERR1\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unsupported command')

        # ERR2 - out of parameter
        with self.assertRaises(aiopjlink.PJLinkERR2) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCD=ERR2\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'out of parameter')

        # NOTE: The following are interpreted as `PJLinkProjectorError`
        # ERR3 - unavailable in the current state
        with self.assertRaises(aiopjlink.PJLinkERR3) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCD=ERR3\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'unavailable in the current state')

        # ERR4 - projector or display failure
        with self.assertRaises(aiopjlink.PJLinkERR4) as err:
            aiopjlink.PJLink._parse_response(
                data='%1ABCD=ERR4\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'projector or display failure')

        # Check error parsing is case insensitive (from projector) and error
        # parsing is handled in the same way too (Postel's law).
        with self.assertRaises(aiopjlink.PJLinkERR4) as err:
            aiopjlink.PJLink._parse_response(
                data='%1abcd=err4\r',
                expect_command='ABCD',
                expect_pjclass='1')
        self.assertEqual(str(err.exception), 'projector or display failure')

        # Check that a respone does not fail if there is a CR right after
        # an equals (e.g. §4.12 of the spec)
        command, param = aiopjlink.PJLink._parse_response(
            data='%1INF2=\r',
            expect_command='INF2',
            expect_pjclass='1')
        self.assertEqual(command, 'INF2')
        self.assertEqual(param, '')


class PowerGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink power control behaves as expected. """

    async def test_state_enum_coercion(self):
        """ Do the enums behave logically. """
        # Power commands.
        self.assertEqual(bool(aiopjlink.Power.ON), True)
        self.assertEqual(bool(aiopjlink.Power.OFF), False)
        self.assertTrue(aiopjlink.Power.ON)
        self.assertFalse(aiopjlink.Power.OFF)

        # Power status (truthy).
        self.assertTrue(aiopjlink.Power.State.ON)
        self.assertTrue(aiopjlink.Power.State.WARMING)

        # Power status (falsy).
        self.assertFalse(aiopjlink.Power.State.COOLING)
        self.assertFalse(aiopjlink.Power.State.OFF)

    async def test_power_get(self):
        """ Get power status. """

        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')
            async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:

                # Power off.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=0\r'):
                    status = await client.power.get()
                    self.assertEqual(status, aiopjlink.Power.OFF)

                # Power on.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=1\r'):
                    status = await client.power.get()
                    self.assertEqual(status, aiopjlink.Power.ON)

                # Power cooling.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=2\r'):
                    status = await client.power.get()
                    self.assertEqual(status, aiopjlink.Power.State.COOLING)

                # Power warming.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=3\r'):
                    status = await client.power.get()
                    self.assertEqual(status, aiopjlink.Power.State.WARMING)

                # Unxpected power result.
                async with server.when(b'%1POWR ?\r', respond_with=b'%1POWR=A\r'):
                    with self.assertRaises(ValueError):
                        await client.power.get()

    async def test_power_set(self):
        """ Set power status. """

        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')
            async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:

                # Power ON
                async with server.when(b'%1POWR 1\r', respond_with=b'%1POWR=OK\r'):
                    await client.power.set(aiopjlink.Power.ON)

                # Power OFF
                async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=OK\r'):
                    await client.power.set(aiopjlink.Power.OFF)

                # Power ON
                async with server.when(b'%2POWR 1\r', respond_with=b'%2POWR=OK\r'):
                    await client.power.set(aiopjlink.Power.ON, pjclass=aiopjlink.PJLink.C2)

                # Power OFF
                async with server.when(b'%2POWR 0\r', respond_with=b'%2POWR=OK\r'):
                    await client.power.set(aiopjlink.Power.OFF, pjclass=aiopjlink.PJLink.C2)

                # Power set out of parameter (test duplicated in test_response_parsing)
                with self.assertRaises(aiopjlink.PJLinkERR2) as err:
                    async with server.when(b'%1POWR 3\r', respond_with=b'%1POWR=ERR2\r'):
                        await client.transmit('POWR', '3', pjclass=aiopjlink.PJLink.C1)
                self.assertEqual(str(err.exception), 'out of parameter')

                # Power set unavailable (test duplicated in test_response_parsing)
                with self.assertRaises(aiopjlink.PJLinkERR3) as err:
                    async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=ERR3\r'):
                        await client.power.set(client.power.OFF)
                self.assertEqual(str(err.exception), 'unavailable in the current state')

                # Power set not possible (test duplicated in test_response_parsing)
                with self.assertRaises(aiopjlink.PJLinkERR4) as err:
                    async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=ERR4\r'):
                        await client.power.set(client.power.OFF)
                self.assertEqual(str(err.exception), 'projector or display failure')

                # Unexpected projector reply to a sensible message.
                with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                    async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=UGH\r'):
                        await client.power.set(aiopjlink.Power.OFF)
                self.assertEqual(str(err.exception), 'expected OK response')

                # Expect enums.
                with self.assertRaises(ValueError) as err:
                    async with server.when(b'%1POWR 1\r', respond_with=b'%1POWR=OK\r'):
                        await client.power.set(True)
                self.assertEqual(str(err.exception), 'True is not a valid Power.State')

                with self.assertRaises(ValueError) as err:
                    async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=OK\r'):
                        await client.power.set('off')
                self.assertEqual(str(err.exception), '\'off\' is not a valid Power.State')

                with self.assertRaises(ValueError) as err:
                    async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=OK\r'):
                        await client.power.set(aiopjlink.Power.State.COOLING)
                self.assertEqual(str(err.exception), 'expected Power.State.ON or Power.State.OFF')

                # Accept PJLink strings as enum values.
                async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=OK\r'):
                    await client.power.set('0')

    async def test_power_shortcuts(self):
        """ Set power status. """

        # Start server with no auth.
        async with mock_tcp_pjlink() as server:
            server.open_and_send(b'PJLINK 0\r')
            async with aiopjlink.PJLink(address='127.0.0.1', password=None) as client:

                # Power ON
                async with server.when(b'%1POWR 1\r', respond_with=b'%1POWR=OK\r'):
                    await client.power.turn_on()

                # Power OFF
                async with server.when(b'%1POWR 0\r', respond_with=b'%1POWR=OK\r'):
                    await client.power.turn_off()


class SourcesGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink input source control and enumeration behaves as expected. """

    async def test_enum(self):
        """ Ensure the enumeration matches the spec. """
        self.assertEqual(aiopjlink.Sources.Mode.RGB.value, '1')
        self.assertEqual(aiopjlink.Sources.Mode.VIDEO.value, '2')
        self.assertEqual(aiopjlink.Sources.Mode.DIGITAL.value, '3')
        self.assertEqual(aiopjlink.Sources.Mode.STORAGE.value, '4')
        self.assertEqual(aiopjlink.Sources.Mode.NETWORK.value, '5')
        self.assertEqual(aiopjlink.Sources.Mode.INTERNAL.value, '6')

    async def test_inpt(self):
        """ Check the INPT instructions work. """
        async with mock_client_server_noauth() as (server, client):

            # Get current source.
            async with server.when(b'%1INPT ?\r', respond_with=b'%1INPT=31\r'):
                mode, index = await client.sources.get()
                self.assertEqual(mode, client.sources.Mode.DIGITAL)
                self.assertEqual(index, '1')

            # Set current source.
            async with server.when(b'%1INPT 21\r', respond_with=b'%1INPT=OK\r'):
                await client.sources.set(aiopjlink.Sources.Mode.VIDEO, '1')

    async def test_inst_class1(self):
        """ Test that available sources can be enumerated. """
        async with mock_client_server_noauth() as (server, client):

            # List available clients
            async with server.when(b'%1INST ?\r', respond_with=b'%1INST=11 31 32 41 52 56\r'):
                sources = await client.sources.available()

                # Length
                self.assertEqual(len(sources), 6)

                # RGB 1
                mode, index = sources[0]
                self.assertEqual(mode, aiopjlink.Sources.Mode.RGB)
                self.assertEqual(index, '1')

                # DIGITAL 1
                mode, index = sources[1]
                self.assertEqual(mode, aiopjlink.Sources.Mode.DIGITAL)
                self.assertEqual(index, '1')

                # DIGITAL 2
                mode, index = sources[2]
                self.assertEqual(mode, aiopjlink.Sources.Mode.DIGITAL)
                self.assertEqual(index, '2')

                # STORAGE 1
                mode, index = sources[3]
                self.assertEqual(mode, aiopjlink.Sources.Mode.STORAGE)
                self.assertEqual(index, '1')

                # NETWORK 2
                mode, index = sources[4]
                self.assertEqual(mode, aiopjlink.Sources.Mode.NETWORK)
                self.assertEqual(index, '2')

                # NETWORK 6
                mode, index = sources[5]
                self.assertEqual(mode, aiopjlink.Sources.Mode.NETWORK)
                self.assertEqual(index, '6')

    async def test_inst_innm_class2(self):
        """ Test that available sources and their names can be enumerated. """
        async with mock_client_server_noauth() as (server, client):

            # List available clients (Class 2) - brief as the Class 1 shares the logic.
            async with server.when(b'%2INST ?\r', respond_with=b'%2INST=11 31 32 41 52 56\r'):
                sources = await client.sources.available(pjclass=aiopjlink.PJClass.TWO)
                self.assertEqual(len(sources), 6)

            # Get the names of a display source (values taken from actual projector output).
            async with server.when(b'%2INNM ?31\r', respond_with=b'%2INNM=DVI-D\r'):
                name = await client.sources.get_source_name(aiopjlink.Sources.Mode.DIGITAL, '1')
                self.assertEqual(name, 'DVI-D')

            # Accept integers (because we are liberal in what we accept).
            async with server.when(b'%2INNM ?31\r', respond_with=b'%2INNM=DVI-D\r'):
                name = await client.sources.get_source_name(aiopjlink.Sources.Mode.DIGITAL, 1)
                self.assertEqual(name, 'DVI-D')

            # Reject long numbers (before hitting the server).
            with self.assertRaises(ValueError) as err:
                name = await client.sources.get_source_name(aiopjlink.Sources.Mode.DIGITAL, 11)
            self.assertEqual(
                str(err.exception),
                'index must be a single character (1-9 for Class 1, and 1-9A-Z for Class 2)'
            )

            # Get the names of available.
            async with server.when(b'%2INST ?\r', respond_with=b'%2INST=11\r'):
                async with server.when(b'%2INNM ?11\r', respond_with=b'%2INNM=Computer\r'):
                    sources = await client.sources.available_with_names()
                    self.assertEqual(len(sources), 1)
                    mode, index, name = sources[0]
                    self.assertEqual(mode, aiopjlink.Sources.Mode.RGB)
                    self.assertEqual(index, '1')
                    self.assertEqual(name, 'Computer')

    async def test_ires(self):
        """ Test that the resolution of the current input can be recieved. """
        async with mock_client_server_noauth() as (server, client):

            # Get resolution.
            async with server.when(b'%2IRES ?\r', respond_with=b'%2IRES=100x200\r'):
                x, y = await client.sources.resolution()
                self.assertEqual(x, 100)
                self.assertEqual(y, 200)

            # Handle spec edge cases.
            with self.assertRaises(aiopjlink.PJLinkProjectorError) as err:
                async with server.when(b'%2IRES ?\r', respond_with=b'%2IRES=-\r'):
                    await client.sources.resolution()
            self.assertEqual(str(err.exception), 'no signal input')

            # Handle spec edge cases.
            with self.assertRaises(aiopjlink.PJLinkProjectorError) as err:
                async with server.when(b'%2IRES ?\r', respond_with=b'%2IRES=*\r'):
                    await client.sources.resolution()
            self.assertEqual(str(err.exception), 'unknown signal')

            # Bad response from the projecetor.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%2IRES ?\r', respond_with=b'%2IRES=\r'):
                    x, y = await client.sources.resolution()
            self.assertEqual(str(err.exception), 'unable to parse resolution')

    async def test_rres(self):
        """ Test that the recommended resolution for the current input can be recieved. """
        async with mock_client_server_noauth() as (server, client):

            # Get resolution.
            async with server.when(b'%2RRES ?\r', respond_with=b'%2RRES=1920x1080\r'):
                x, y = await client.sources.recommended_resolution()
                self.assertEqual(x, 1920)
                self.assertEqual(y, 1080)

            # Bad response from the projecetor.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%2RRES ?\r', respond_with=b'%2RRES=\r'):
                    x, y = await client.sources.recommended_resolution()
            self.assertEqual(str(err.exception), 'unable to parse resolution')


class MuteGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink mute controls behave as expected. """

    async def test_status(self):
        """ Mute status is acquired OK. """
        async with mock_client_server_noauth() as (server, client):

            # Video only muted.
            async with server.when(b'%1AVMT ?\r', respond_with=b'%1AVMT=11\r'):
                video, audio = await client.mute.status()
                self.assertEqual(video, True)
                self.assertEqual(audio, False)

            # Audio only muted.
            async with server.when(b'%1AVMT ?\r', respond_with=b'%1AVMT=21\r'):
                video, audio = await client.mute.status()
                self.assertEqual(video, False)
                self.assertEqual(audio, True)

            # Nothing muted.
            async with server.when(b'%1AVMT ?\r', respond_with=b'%1AVMT=30\r'):
                video, audio = await client.mute.status()
                self.assertEqual(video, False)
                self.assertEqual(audio, False)

            # Both audio and video muted.
            async with server.when(b'%1AVMT ?\r', respond_with=b'%1AVMT=31\r'):
                video, audio = await client.mute.status()
                self.assertEqual(video, True)
                self.assertEqual(audio, True)

            # Bad information from projector.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%1AVMT ?\r', respond_with=b'%1AVMT=41\r'):
                    video, audio = await client.mute.status()
            self.assertEqual(str(err.exception), 'unexpected mute response')

    async def test_control_api(self):
        """ Mute status can be set OK using expressive methods. """
        async with mock_client_server_noauth() as (server, client):

            # VIDEO
            # %1AVMT 11	%1AVMT=OK	blanking on
            async with server.when(b'%1AVMT 11\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.video(True)
            # %1AVMT 10	%1AVMT=OK	blanking off
            async with server.when(b'%1AVMT 10\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.video(False)

            # AUDIO
            # %1AVMT 21	%1AVMT=OK	audio muting on
            async with server.when(b'%1AVMT 21\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.audio(True)
            # %1AVMT 20	%1AVMT=OK	audio muting off
            async with server.when(b'%1AVMT 20\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.audio(False)

            # BOTH AT ONCE
            # %1AVMT 31	%1AVMT=OK	blanking on and audio muting on
            async with server.when(b'%1AVMT 31\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.both(True)

            # %1AVMT 30	%1AVMT=OK	blanking on and audio muting off
            async with server.when(b'%1AVMT 30\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.both(False)

    async def test_set(self):
        """ Mute status can be set OK using the shortcut set method. """
        async with mock_client_server_noauth() as (server, client):

            # Fully specified (3x condition).
            async with server.when(b'%1AVMT 31\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.set(True, True)
            async with server.when(b'%1AVMT 30\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.set(False, False)

            # Fully specified (2x instructions)
            async with server.when(b'%1AVMT 11\r', respond_with=b'%1AVMT=OK\r'):
                async with server.when(b'%1AVMT 20\r', respond_with=b'%1AVMT=OK\r'):
                    await client.mute.set(True, False)

            async with server.when(b'%1AVMT 10\r', respond_with=b'%1AVMT=OK\r'):
                async with server.when(b'%1AVMT 21\r', respond_with=b'%1AVMT=OK\r'):
                    await client.mute.set(False, True)

            # Partially specified (for audio) (1x instruction)
            async with server.when(b'%1AVMT 21\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.set(video=None, audio=True)
            async with server.when(b'%1AVMT 20\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.set(video=None, audio=False)

            # Partially specified (for video) (1x instruction)
            async with server.when(b'%1AVMT 11\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.set(video=True, audio=None)
            async with server.when(b'%1AVMT 10\r', respond_with=b'%1AVMT=OK\r'):
                await client.mute.set(video=False, audio=None)

            # Nothing specified - so no messages sent.
            with self.assertRaises(NoClientMessage):
                async with server.when(b'%1XXXX 11\r', respond_with=b'NOTHING\r'):
                    await client.mute.set(None, None)


class ErrorGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink error reporting behaves as expected. """

    async def test_query(self):
        """ Error status parses results correctly. """
        async with mock_client_server_noauth() as (server, client):

            # No errors!
            async with server.when(b'%1ERST ?\r', respond_with=b'%1ERST=000000\r'):
                errors = await client.errors.query()
                self.assertEqual(errors, {
                    aiopjlink.Errors.Category.FAN: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.LAMP: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.TEMP: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.COVER: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.FILTER: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.OTHER: aiopjlink.Errors.Level.OK,
                })

            # Two warnings and one failure.
            async with server.when(b'%1ERST ?\r', respond_with=b'%1ERST=101020\r'):
                errors = await client.errors.query()
                self.assertEqual(errors, {
                    aiopjlink.Errors.Category.FAN: aiopjlink.Errors.Level.WARN,
                    aiopjlink.Errors.Category.LAMP: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.TEMP: aiopjlink.Errors.Level.WARN,
                    aiopjlink.Errors.Category.COVER: aiopjlink.Errors.Level.OK,
                    aiopjlink.Errors.Category.FILTER: aiopjlink.Errors.Level.ERROR,
                    aiopjlink.Errors.Category.OTHER: aiopjlink.Errors.Level.OK,
                })

            # Bad error from the projector - too many errors.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%1ERST ?\r', respond_with=b'%1ERST=0000000\r'):
                    errors = await client.errors.query()
            self.assertEqual(str(err.exception), 'unexpected number of error types reported')

            # Bad error from the projector - unknown error type.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%1ERST ?\r', respond_with=b'%1ERST=003000\r'):
                    errors = await client.errors.query()
            self.assertEqual(str(err.exception), 'unknown error level')


class FilterGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink filter reports correctly. """

    async def test_hours(self):
        """ Get the correct filter hours. """
        async with mock_client_server_noauth() as (server, client):

            # Valid filter.
            async with server.when(b'%2FILT ?\r', respond_with=b'%2FILT=100\r'):
                hours = await client.filter.hours()
                self.assertEqual(hours, 100)

            # No filter handled correctly.
            with self.assertRaises(aiopjlink.PJLinkERR1) as err:
                async with server.when(b'%2FILT ?\r', respond_with=b'%2FILT=ERR1\r'):
                    hours = await client.filter.hours()
            self.assertEqual(str(err.exception), 'no filter')

    async def test_replacement_model(self):
        """ Get a list of replacement filter models. """
        async with mock_client_server_noauth() as (server, client):

            # Once replacement.
            async with server.when(b'%2RFIL ?\r', respond_with=b'%2RFIL=SampleFilter\r'):
                models = await client.filter.replacement_models()
                self.assertEqual(len(models), 1)
                self.assertEqual(models[0], 'SampleFilter')

            # Two replacements.
            async with server.when(b'%2RFIL ?\r', respond_with=b'%2RFIL=SampleFilter ELPAF46\r'):
                models = await client.filter.replacement_models()
                self.assertEqual(len(models), 2)
                self.assertEqual(models[0], 'SampleFilter')
                self.assertEqual(models[1], 'ELPAF46')

            # No replacements.
            async with server.when(b'%2RFIL ?\r', respond_with=b'%2RFIL=\r'):
                models = await client.filter.replacement_models()
                self.assertEqual(len(models), 0)


class LampGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink lamps report correctly. """

    async def test_status(self):
        """ Lamp status parses correctly for common values. """
        async with mock_client_server_noauth() as (server, client):

            # Two lamps (from PJLink test excel sheet)
            async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=8253 1 13442 1\r'):
                lamps = await client.lamps.status()
                self.assertEqual(len(lamps), 2)

                hours, state = lamps[0]
                self.assertEqual(hours, 8253)
                self.assertEqual(state, aiopjlink.Lamp.State.ON)

                hours, state = lamps[1]
                self.assertEqual(hours, 13442)
                self.assertEqual(state, aiopjlink.Lamp.State.ON)

            # One lamp off
            async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=8253 0\r'):
                lamps = await client.lamps.status()
                self.assertEqual(len(lamps), 1)
                hours, state = lamps[0]
                self.assertEqual(hours, 8253)
                self.assertEqual(state, aiopjlink.Lamp.State.OFF)

            # No lamps in the projector.
            with self.assertRaises(aiopjlink.PJLinkERR1):
                async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=ERR1\r'):
                    lamps = await client.lamps.status()

            # Unparsable feedback from projector.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=1 0 10\r'):
                    lamps = await client.lamps.status()
            self.assertEqual(str(err.exception), 'unparsable lamp status')

            # Bad feedback from the projector.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=\r'):
                    lamps = await client.lamps.status()
            self.assertEqual(str(err.exception), 'unparsable lamp status')

            # Bad feedback from the projector.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=100 3\r'):
                    lamps = await client.lamps.status()
            self.assertEqual(str(err.exception), 'unparsable lamp status')

    async def test_hours(self):
        """ The hours accelerator gets the correct lamp hours. """
        async with mock_client_server_noauth() as (server, client):

            # Two lamps (from PJLink test excel sheet)
            async with server.when(b'%1LAMP ?\r', respond_with=b'%1LAMP=8253 1 13442 1\r'):
                hours = await client.lamps.hours()
                self.assertEqual(hours, 8253)

    async def test_replacement_model(self):
        """ Get a list of replacement lamp models. """
        async with mock_client_server_noauth() as (server, client):

            # Once replacement.
            async with server.when(b'%2RLMP ?\r', respond_with=b'%2RLMP=SampleLamp\r'):
                models = await client.lamps.replacement_models()
                self.assertEqual(len(models), 1)
                self.assertEqual(models[0], 'SampleLamp')

            # Two replacements.
            async with server.when(b'%2RLMP ?\r', respond_with=b'%2RLMP=SampleLamp RC11\r'):
                models = await client.lamps.replacement_models()
                self.assertEqual(len(models), 2)
                self.assertEqual(models[0], 'SampleLamp')
                self.assertEqual(models[1], 'RC11')

            # No replacements.
            async with server.when(b'%2RLMP ?\r', respond_with=b'%2RLMP=\r'):
                models = await client.lamps.replacement_models()
                self.assertEqual(len(models), 0)


class FreezeGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink freeze behavour activates and reports correctly. """

    async def test_set(self):
        """ Screen freezes and unfreezes. """
        async with mock_client_server_noauth() as (server, client):

            # Expected.
            async with server.when(b'%2FREZ 1\r', respond_with=b'%2FREZ=OK\r'):
                await client.freeze.set(True)
            async with server.when(b'%2FREZ 0\r', respond_with=b'%2FREZ=OK\r'):
                await client.freeze.set(False)

            # Unexpected.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%2FREZ 1\r', respond_with=b'%2FREZ=UGH\r'):
                    await client.freeze.set(True)
            self.assertEqual(str(err.exception), 'expected OK response')

    async def test_get(self):
        """ Getting the current freeze state. """
        async with mock_client_server_noauth() as (server, client):

            # Expected.
            async with server.when(b'%2FREZ ?\r', respond_with=b'%2FREZ=0\r'):
                self.assertEqual(await client.freeze.get(), False)
            async with server.when(b'%2FREZ ?\r', respond_with=b'%2FREZ=1\r'):
                self.assertEqual(await client.freeze.get(), True)

            # Unexpected.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%2FREZ ?\r', respond_with=b'%2FREZ=2\r'):
                    await client.freeze.get()
            self.assertEqual(str(err.exception), 'unexpected freeze state')


class VolumeGroups(unittest.IsolatedAsyncioTestCase):
    """ PJLink volume increments and decrements correctly. """

    async def test_change_volume(self):
        """ Volume turns up and down for speaker and microphone. """
        async with mock_client_server_noauth() as (server, client):

            # Turn up mic.
            async with server.when(b'%2MVOL 1\r', respond_with=b'%2MVOL=OK\r'):
                await client.microphone.turn_up()

            # Turn up speaker.
            async with server.when(b'%2SVOL 1\r', respond_with=b'%2SVOL=OK\r'):
                await client.speaker.turn_up()

            # Turn down mic.
            async with server.when(b'%2MVOL 0\r', respond_with=b'%2MVOL=OK\r'):
                await client.microphone.turn_down()

            # Turn down speaker.
            async with server.when(b'%2SVOL 0\r', respond_with=b'%2SVOL=OK\r'):
                await client.speaker.turn_down()

            # Unexpected.
            with self.assertRaises(aiopjlink.PJLinkUnexpectedResponseParameter) as err:
                async with server.when(b'%2SVOL 1\r', respond_with=b'%2SVOL=2\r'):
                    await client.speaker.turn_up()
            self.assertEqual(str(err.exception), 'expected OK response')


class InfoGroup(unittest.IsolatedAsyncioTestCase):
    """ PJLink projector information queries. """

    async def test_software_version(self):
        """ Get the software version number string. """
        async with mock_client_server_noauth() as (server, client):

            # Version provided.
            async with server.when(b'%2SVER ?\r', respond_with=b'%2SVER=24011273HQWWV105\r'):
                version = await client.info.software_version()
                self.assertEqual(version, '24011273HQWWV105')

            # No version provided.
            async with server.when(b'%2SVER ?\r', respond_with=b'%2SVER=\r'):
                version = await client.info.software_version()
                self.assertEqual(version, '')

    async def test_serial_number(self):
        """ Get the device serial number string. """
        async with mock_client_server_noauth() as (server, client):

            # Version provided.
            async with server.when(b'%2SNUM ?\r', respond_with=b'%2SNUM=XA3C2400119\r'):
                version = await client.info.serial_number()
                self.assertEqual(version, 'XA3C2400119')

            # No version provided.
            async with server.when(b'%2SNUM ?\r', respond_with=b'%2SNUM=\r'):
                version = await client.info.serial_number()
                self.assertEqual(version, '')

    async def test_pjlink_class(self):
        """ Query the PJLink class support. """
        async with mock_client_server_noauth() as (server, client):

            # Class 1 query (default).
            async with server.when(b'%1CLSS ?\r', respond_with=b'%1CLSS=1\r'):
                self.assertEqual(await client.info.pjlink_class(), aiopjlink.PJClass.ONE)

            # Class 2 support from class 1 query (default).
            async with server.when(b'%1CLSS ?\r', respond_with=b'%1CLSS=2\r'):
                self.assertEqual(await client.info.pjlink_class(), aiopjlink.PJClass.TWO)

            # Class 2 support from class 2 query (default).
            async with server.when(b'%2CLSS ?\r', respond_with=b'%2CLSS=2\r'):
                response = await client.info.pjlink_class(pjclass=aiopjlink.PJClass.TWO)
                self.assertEqual(response, aiopjlink.PJClass.TWO)

    async def test_info_other(self):
        """ Query the "other" info. """
        async with mock_client_server_noauth() as (server, client):

            # Test from spec excel sheet.
            async with server.when(b'%1INFO ?\r', respond_with=b'%1INFO=PJLink\r'):
                self.assertEqual(await client.info.other(), 'PJLink')

            # Test actual projector response.
            async with server.when(b'%1INFO ?\r', respond_with=b'%1INFO=105.105.---\r'):
                self.assertEqual(await client.info.other(), '105.105.---')

            # Test no extra info.
            async with server.when(b'%1INFO ?\r', respond_with=b'%1INFO=\r'):
                self.assertEqual(await client.info.other(), '')

    async def test_product_name(self):
        """ Query the product name information. """
        async with mock_client_server_noauth() as (server, client):

            # Test from actual projector.
            async with server.when(b'%1INF2 ?\r', respond_with=b'%1INF2=EPSON PU1007B/PU1007W\r'):
                self.assertEqual(await client.info.product_name(), 'EPSON PU1007B/PU1007W')

            # Test no extra info.
            async with server.when(b'%1INF2 ?\r', respond_with=b'%1INF2=\r'):
                self.assertEqual(await client.info.product_name(), '')

    async def test_manufacturer_name(self):
        """ Query the manufacturer name information. """
        async with mock_client_server_noauth() as (server, client):

            # Test from actual projector.
            async with server.when(b'%1INF1 ?\r', respond_with=b'%1INF1=EPSON\r'):
                self.assertEqual(await client.info.manufacturer_name(), 'EPSON')

            # Test no extra info.
            async with server.when(b'%1INF1 ?\r', respond_with=b'%1INF1=\r'):
                self.assertEqual(await client.info.manufacturer_name(), '')

    async def test_projector_name(self):
        """ Query the projector name information. """
        async with mock_client_server_noauth() as (server, client):

            # Test from actual projector.
            async with server.when(b'%1NAME ?\r', respond_with=b'%1NAME=EBB13648\r'):
                self.assertEqual(await client.info.projector_name(), 'EBB13648')

            # Test no extra info.
            async with server.when(b'%1NAME ?\r', respond_with=b'%1NAME=\r'):
                self.assertEqual(await client.info.projector_name(), '')
